/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/SeparatoG.h>
#include <Xm/ToggleB.h>

static char SccsID[] = "@(#)alloc_tree.c	1.2\t11/25/92";            /* sccs keyword */

static void setTreeHints();
static void createTreeViewPort();
static void createTreeScrollbars();
static void setTreeScrollbars();
static void setTree_hScrollbar();
static void setTree_vScrollbar();

void SelectTreeWnd();
void redrawTreeWindow();
void SelectDirTreeNode();
void confTreeWindow();
void closeTree();
void treeVscroll();
void treeHscroll();

static void treeVmove();
static void treeHmove();

dirWindow *lookifDisplayed();

dirNode *makeDirNodeStruct();

extern treeWindow *curTreeWnd;

/* makeTreeWindow: ************************************************************
 * allocate and initialise the structure that holds information pertaining
 * to a directory tree window 
 *****************************************************************************/

treeWindow *makeTreeWindow(wnd)
     dirWindow *wnd;
{
  Widget w;
  treeWindow *tw;
  Arg args[15];
  XmFontList fl;
  int cnt;
  static XtTranslations ptt = NULL;

  static char tree_translations[] = 
    " <Map>		: setTreeHints()";

  static XtActionsRec tree_actions[] = {
    { "setTreeHints", setTreeHints },
  };

  if (! ptt) {
    ptt = XtParseTranslationTable(tree_translations);
    XtAddActions(tree_actions, 1);
  }
  
  tw = (treeWindow *) Malloc(sizeof(treeWindow));
  tw->nodes = (dirNode *) NULL;
  tw->path = strCopy(wnd->path);
  tw->wnd = wnd;

  /* top level shell */
  cnt = 0;
  XtSetArg(args[cnt], XmNborderWidth, BORDERWIDTH); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
  XtSetArg(args[cnt], XmNiconName, wnd->path); cnt++;
  tw->shell = XtCreatePopupShell("tree_shell", topLevelShellWidgetClass, 
				 		tw->wnd->fmTop, args, cnt);
  /* top-level child */
  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, tw); cnt++;
  tw->w = XtCreateManagedWidget("tree_box", 
				xmFormWidgetClass, tw->shell, args, cnt);
  XtOverrideTranslations(tw->w, ptt);

  /* close box */
  cnt = 0;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelType, XmPIXMAP); cnt++;
  XtSetArg(args[cnt], XmNlabelPixmap, closeBoxPixmap); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  tw->closeBox  = 
    XtCreateManagedWidget("close_box", xmPushButtonWidgetClass,
				      tw->w, args, cnt);
  XtAddCallback(tw->closeBox, XmNactivateCallback, closeTree, tw->wnd); 

  /*create title-bar */
  cnt = 0;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNrightWidget, tw->closeBox); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
  tw->title = XtCreateManagedWidget("title", xmLabelWidgetClass,
				     			tw->w, args, cnt);
  setTitle(tw->title, tw->path);
  
  /* separator */
  cnt = 0;
  XtSetArg(args[cnt], XmNseparatorType, XmSHADOW_ETCHED_IN); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, SEP_THICKNESS); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, tw->title); cnt++;
  tw->separator = XtCreateManagedWidget("sep", 
			      xmSeparatorGadgetClass, tw->w, args, cnt);

  createTreeScrollbars(tw);
  createTreeViewPort(tw);

  tw->cur_col = tw->cur_row = 0;

  /* create a dummy node to be able to retrieve the font resource */

  w = XtCreateManagedWidget("node", xmToggleButtonWidgetClass, 
				    		tw->plane, NULL, 0);
  /* get font resource from buttons */
  cnt = 0;
  XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
  XtGetValues(w, args, cnt);

  if (! fl)
    tw->font = exBaseFont(XtDisplay(Head->root));
  else
    tw->font = getFont(fl);

  XtDestroyWidget(w);

  return(tw);
}

/* createTreeViewPort: *******************************************************
 * create the viewport window that defines what buttons are visible
 ****************************************************************************/

static void createTreeViewPort(tw)
     treeWindow *tw;
{
  Arg args[15];
  int cnt;
  static XtTranslations ptt = NULL, ptt1 = NULL;

  static char treewnd_translations[] = 
   "<EnterWindow>	: SelectTreeWindow() \n \
    <LeaveWindow>	: SelectTreeWindow() \n \
    <Expose>		: redrawTreeWindow()";

  static char treework_translations[] = 
   "<ConfigureNotify>	: confTreeWindow()";

  static XtActionsRec treewnd_actions[] = {
    { "SelectTreeWindow", SelectTreeWnd},
    { "redrawTreeWindow", redrawTreeWindow},
    { "confTreeWindow", confTreeWindow}
  };

  static XtActionsRec treework_actions[] = {
    { "confTreeWindow", confTreeWindow}
  };

  if (! ptt) {
    XtAddActions(treewnd_actions, 2);
    XtAddActions(treework_actions, 1);
    ptt = XtParseTranslationTable(treewnd_translations);
    ptt1 = XtParseTranslationTable(treework_translations);
  }

  /* create view-port */
  cnt = 0;
  if (tw->v_scrollb) {
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, tw->v_scrollb); cnt++;
  } else {
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  }

  if (tw->h_scrollb) {
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNbottomWidget, tw->h_scrollb); cnt++;
  } else {
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  }

  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, tw->separator); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNuserData, tw); cnt++;
  tw->work = XtCreateManagedWidget("view_port", 
				  xmDrawingAreaWidgetClass, tw->w, args, cnt);
  XtOverrideTranslations(tw->work, ptt1);

  /* create tree plane */
  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, tw); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNx, 0); cnt++;
  XtSetArg(args[cnt], XmNy, 0); cnt++;
  XtSetArg(args[cnt], XmNwidth, 10000); cnt++;
  XtSetArg(args[cnt], XmNheight, 10000); cnt++;
  
  tw->plane = XtCreateManagedWidget("plane", xmDrawingAreaWidgetClass,
				    tw->work, args, cnt);

  XtOverrideTranslations(tw->plane, ptt);
}

/* setTreeScrollbars: *******************************************************
 * set scrollbar values
 ****************************************************************************/
static void setTreeScrollbars(tw, sb_width, width, height)
     treeWindow *tw;
     int sb_width, width, height;
{
  Arg args[15];
  int cnt;
  int size, inc, pageinc, bh, bw, h;
  
  bh = DIRBUTTONHEIGHT(tw->font);
  bw = DIRBUTTONWIDTH(tw->font);
  
  /* set left (vertical) scrollbar */
  if (! (h = ROUNDOFF(height, (bh + DIRBUTTONVGAP)))) {
    tw->flags &= ~VSCROLL;
    size = height;
  } else {
    size = (int) (float) h / (float) tw->rows * (float) height;
    tw->flags |= VSCROLL;
  }

  inc = (size == height) ? 1 : 
    (int)((float)(height - size) / (float)(tw->rows - h));

  pageinc = MAX(1, MIN(4, tw->rows - DEFNODESPERROW));
  
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, sb_width); cnt++;
  XtSetArg(args[cnt], XmNsliderSize, size); cnt++;
  XtSetArg(args[cnt], XmNminimum, 0); cnt++;
  XtSetArg(args[cnt], XmNmaximum, height); cnt++;
  XtSetArg(args[cnt], XmNincrement, inc); cnt++;
  XtSetArg(args[cnt], XmNpageIncrement, 
	MAX(1, (int)((float)(height - size) / (float) pageinc))); cnt++;
  XtSetArg(args[cnt], XmNvalue, 0); cnt++;
  XtSetValues(tw->v_scrollb, args, cnt);
  
  /* set bottom (horizontal) scrollbar */
  if (! (h = ROUNDOFF(width, (bw + DIRBUTTONHGAP)))) {
    tw->flags &= ~HSCROLL;
    size = width;
  } else {
    size = (int) (float) h / (float) tw->cols * (float) width;
    tw->flags |= HSCROLL;
  }

  inc = (size == width) ? 1 : 
    	(int) ((float)(width - size) / (float)(tw->cols - h));
  pageinc = MAX(1, MIN(4, tw->cols - DEFNODESPERLINE));
  
  cnt = 0;
  XtSetArg(args[cnt], XmNheight, sb_width); cnt++;
  XtSetArg(args[cnt], XmNsliderSize, size); cnt++;
  XtSetArg(args[cnt], XmNminimum, 0); cnt++;
  XtSetArg(args[cnt], XmNmaximum, width); cnt++;
  XtSetArg(args[cnt], XmNvalue, 0); cnt++;
  XtSetArg(args[cnt], XmNincrement, inc); cnt++;
  XtSetArg(args[cnt], XmNpageIncrement, 
	MAX(1, (int)((float)(width - size) / (float) pageinc))); cnt++;
  XtSetValues(tw->h_scrollb, args, cnt);
}

/* createTreeScrollbars: *****************************************************
 * create the tree window scrollbars
 ****************************************************************************/
static void createTreeScrollbars(tw)
     treeWindow *tw;
{
  Arg args[15];
  int cnt;
  
  /* create left (vertical ) scrollbar */
  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, tw->separator); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNorientation, XmVERTICAL); cnt++;
  tw->v_scrollb = XtCreateManagedWidget("vScrollBar", 
			xmScrollBarWidgetClass, tw->w, args, cnt);
  XtAddCallback(tw->v_scrollb, XmNdecrementCallback, treeVscroll, tw);
  XtAddCallback(tw->v_scrollb, XmNincrementCallback, treeVscroll, tw);
  XtAddCallback(tw->v_scrollb, XmNpageDecrementCallback, treeVscroll, tw);
  XtAddCallback(tw->v_scrollb, XmNpageIncrementCallback, treeVscroll, tw);
  XtAddCallback(tw->v_scrollb, XmNdragCallback, treeVscroll, tw);
  
  /* create bottom (horizontal) scrollbar */
  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, tw->v_scrollb); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;
  tw->h_scrollb = XtCreateManagedWidget("hScrollBar", 
			xmScrollBarWidgetClass, tw->w, args, cnt);
  XtAddCallback(tw->h_scrollb, XmNdecrementCallback, treeHscroll, tw);
  XtAddCallback(tw->h_scrollb, XmNincrementCallback, treeHscroll, tw);
  XtAddCallback(tw->h_scrollb, XmNpageDecrementCallback, treeHscroll, tw);
  XtAddCallback(tw->h_scrollb, XmNpageIncrementCallback, treeHscroll, tw);
  XtAddCallback(tw->h_scrollb, XmNdragCallback, treeHscroll, tw);
}

/* setTreeHints: *************************************************************
 * set hints fro the window manager to allow resizing only in reasonable steps 
 ****************************************************************************/

static void setTreeHints(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  treeWindow *tw;
  XSizeHints size_hints;
  int sb_width, th, sh;
  Arg args[1];
  int cnt;
  
  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &tw); cnt++;

  XtGetValues(w, args, cnt);

  if (! tw)
    return;

  sb_width = exFontHeight(tw->font);
  th = titleHeight(tw->title);
  sh = sepHeight(tw->separator);

  size_hints.flags = PMinSize | PResizeInc | PMaxSize;

  size_hints.min_width = DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP + sb_width;
  size_hints.min_height = 
    sb_width + th + sh + DIRBUTTONHEIGHT(tw->font) + DIRBUTTONVGAP;

  size_hints.max_width = 
    MIN(10, tw->cols) * (DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP) + sb_width;
  size_hints.max_height = tw->rows * (DIRBUTTONHEIGHT(tw->font) + 
		      			DIRBUTTONVGAP) + sb_width + th + sh;

  size_hints.width_inc  = DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP;
  size_hints.height_inc = DIRBUTTONHEIGHT(tw->font) + DIRBUTTONVGAP;

  XSetNormalHints(XtDisplay(tw->shell), XtWindow(tw->shell), &size_hints);
}

/* createDirNodes: ***********************************************************
 * create the correspoding buttons
 ****************************************************************************/

createDirNodes(tw)
     treeWindow *tw;
{
  register dirNode *d;
  int cnt;
  char *pt;
  char t[NODELEN + 1];
  char *memset();
  char *tail();
  Arg args[15];
  int height, width, fullHeight, fullWidth, sb_width, th, sh;

  for (d = tw->nodes; d; d = d->next) {
    memset(t, ' ', NODELEN);
    pt = tail(d->name);
    bcopy(pt, t, MIN(strlen(pt), NODELEN));
    t[NODELEN] = EOS;

    cnt = 0;
    XtSetArg(args[cnt], XmNx, d->x); cnt++;
    XtSetArg(args[cnt], XmNy, d->y); cnt++;
    XtSetArg(args[cnt], XmNspacing, 0); cnt++;
    XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
    XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNshadowThickness, SHADOW_THICKNESS); cnt++;
    XtSetArg(args[cnt], XmNindicatorOn, False); cnt++;

    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreate(t, XmSTRING_DEFAULT_CHARSET)); cnt++; 
    
    if (lookifDisplayed(d->name)) {
      XtSetArg(args[cnt], XmNforeground, fm_white_pixel); cnt++; 
      XtSetArg(args[cnt], XmNbackground, fm_black_pixel); cnt++; 
      d->flags |= NODE_DISPLAYED;
    } else {
      XtSetArg(args[cnt], XmNforeground, fm_black_pixel); cnt++; 
      XtSetArg(args[cnt], XmNbackground, fm_white_pixel); cnt++; 
    }
    
    d->node = XtCreateManagedWidget("node", xmToggleButtonWidgetClass, 
				    tw->plane, args, cnt);
    
    XtAddCallback(d->node, XmNvalueChangedCallback, SelectDirTreeNode, d);

  }
  
  /* calculate some dimensions that are font dependant */
  sb_width = exFontHeight(tw->font);

  width = MIN(DEFNODESPERLINE, tw->cols) * 
                              (DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP);
  height = MIN(DEFNODESPERROW, tw->rows) * 
    		              (DIRBUTTONHEIGHT(tw->font)+ DIRBUTTONVGAP);

  setTreeScrollbars(tw, sb_width, width, height);

  /* adjust window with respect to button font */
  fullWidth  = tw->v_scrollb ? width  + sb_width : width;
  fullHeight = tw->h_scrollb ? height + sb_width : height;

  th = titleHeight(tw->title);
  sh = sepHeight(tw->separator);
  fullHeight += sh + th;

  /* adjust main window */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, fullWidth); cnt++;
  XtSetArg(args[cnt], XmNheight, fullHeight); cnt++;
  XtSetValues(tw->w, args, cnt);
}

/* deleteTreeWindow: **********************************************************
 * delete the tree window of the given directory window 
 *****************************************************************************/

deleteTreeWindow(wnd)
     dirWindow *wnd;
{
  treeWindow *tw = wnd->tree;
  dirNode *n, *p;
  
  if (! tw)
    return;

  free(tw->path);
  XtDestroyWidget(tw->shell);

  for (n = tw->nodes; n != (dirNode *) NULL; ) {
    free(n->name);
    p = n;
    n = n->next;
    free(p);
  }

  free(tw);
  wnd->tree = (treeWindow *) NULL;
}

/* SelectDirTreeNode: *******************************************************
 * 
 ****************************************************************************/
/* ARGSUSED */
void SelectDirTreeNode(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  register dirNode *b = (dirNode *) client_data;
  XmToggleButtonCallbackStruct *cb; 
  register dirWindow *dw;
  XButtonEvent *e;
  Arg args[5];
  int cnt;

  cb = (XmToggleButtonCallbackStruct *) call_data;
  e = (XButtonEvent *) cb->event;

  if (e->type != ButtonRelease)		/* to be shure */
    return;

  if (e->button != Button1)		/* not left button */
    return;
  
  cnt = 0;
  if (! (b->flags & NODE_DISPLAYED)) {
    if (b->flags & NODE_SELECTED) {
      b->flags &= ~NODE_SELECTED;
      XtSetArg(args[cnt], XmNforeground, fm_black_pixel); cnt++; 
      XtSetArg(args[cnt], XmNbackground, fm_white_pixel); cnt++; 
    } else {
      b->flags |= NODE_SELECTED;
      XtSetArg(args[cnt], XmNforeground, fm_white_pixel); cnt++; 
      XtSetArg(args[cnt], XmNbackground, fm_black_pixel); cnt++; 
    }
    XtSetValues(b->node, args, cnt);
  }
  if (dw = lookifDisplayed(b->name)) {	/* already displayed */
    XWarpPointer(XtDisplay(Head->root), None, XtWindow(dw->workArea), 
		 					0, 0, 0, 0, 10, 10);
  } else {			    	/* create new dir-window */
    listDir(curTreeWnd->wnd, b->name);
  }
  b->flags |= NODE_DISPLAYED;
}

/* treeVscroll: **************************************************************
 * callback function that is called when an event occurs in the verical 
 * scrollbar of a tree window.
 ****************************************************************************/
/* ARGSUSED */
void treeVscroll(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  treeWindow *tw = (treeWindow *) client_data;
  XmScrollBarCallbackStruct *cb = (XmScrollBarCallbackStruct *) call_data;
  int npr, size, max, cnt, pageinc;
  int oldtop = tw->cur_row;
  Arg args[2];

  if (! (tw->flags & VSCROLL))		/* scrollbar not active */
      return;

  switch (cb->reason) {
  case XmCR_DECREMENT:
    if (tw->cur_row > 0)
      tw->cur_row--;
    break;
  case XmCR_INCREMENT:
    npr = nodesPerRow(tw);
    if (tw->cur_row + npr < tw->rows)
      tw->cur_row++;
    break;
  case XmCR_PAGE_DECREMENT:
    npr = nodesPerRow(tw);
    pageinc =  MIN(4, tw->cols - DEFNODESPERLINE);
    if (tw->cur_row - (npr / pageinc) > 0)
      tw->cur_row -= (npr / pageinc);
    else
      tw->cur_row = 0;
    break;
  case XmCR_PAGE_INCREMENT:
    npr = nodesPerRow(tw);
    pageinc =  MIN(4, tw->cols - DEFNODESPERLINE);
    if (tw->cur_row + (npr / pageinc) + npr < tw->rows)
      tw->cur_row += (npr / pageinc);
    else 
      tw->cur_row = tw->rows - npr;
    break;
  case XmCR_DRAG:
    npr = nodesPerRow(tw);

    cnt = 0;
    XtSetArg(args[cnt], XmNmaximum, &max); cnt++;
    XtSetArg(args[cnt], XmNsliderSize, &size); cnt++;
    XtGetValues(tw->v_scrollb, args, cnt);
    
    tw->cur_row = (int)((float) cb->value / (float)(max - size) * 
						(float)(tw->rows - npr));
    
    if (tw->cur_row > tw->rows - npr)
      tw->cur_row = tw->rows - npr;
    break;
  }

  if (oldtop != tw->cur_row) 				/* nothing changed */
    treeVmove(tw);
}

/* treeVmove: ***************************************************************
 * move the button plane so that the current row appears at the top 
 ****************************************************************************/

static void treeVmove(tw)
     treeWindow *tw;
{    
  XtMoveWidget(tw->plane, 
	     -tw->cur_col * (DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP),
	     -tw->cur_row * (DIRBUTTONHEIGHT(tw->font) + DIRBUTTONVGAP));
 
  XDrawSegments(XtDisplay(Head->root), XtWindow(tw->plane), Head->fm_gc, 
		  			tw->connections, tw->conCnt);
}

/* treeHscroll: **************************************************************
 * callback function that is called when an event occurs in the verical 
 * scrollbar of a tree window.
 ****************************************************************************/
/*ARGSUSED*/
void treeHscroll(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  treeWindow *tw = (treeWindow *) client_data;
  XmScrollBarCallbackStruct *cb = (XmScrollBarCallbackStruct *) call_data;
  int npl, max, size, cnt, pageinc;
  int oldmargin = tw->cur_col;
  Arg args[2];
  
    
  if (! (tw->flags & HSCROLL))		/* scrollbar not active */
      return;

  switch (cb->reason) {
  case XmCR_DECREMENT:
    if (tw->cur_col > 0)
      tw->cur_col--;
    break;
  case XmCR_INCREMENT:
    npl = nodesPerLine(tw);
    if (tw->cur_col + npl < tw->cols)
      tw->cur_col++;
    break;
  case XmCR_PAGE_DECREMENT:
    pageinc =  MIN(4, tw->cols - DEFNODESPERLINE);
    npl = nodesPerLine(tw);
    if (tw->cur_col - (npl / pageinc) > 0)
      tw->cur_col -= (npl / pageinc);
    else 
      tw->cur_col = 0;
    break;
  case XmCR_PAGE_INCREMENT:
    pageinc = MIN(4, tw->cols - DEFNODESPERLINE);
    npl = nodesPerLine(tw);
    if (tw->cur_col + tw->cols + (npl / pageinc) < npl)
      tw->cur_col += (npl / pageinc);
    else
      tw->cur_col = tw->cols - npl;
    break;
  case XmCR_DRAG:
    npl = nodesPerLine(tw);

    cnt = 0;
    XtSetArg(args[cnt], XmNmaximum, &max); cnt++;
    XtSetArg(args[cnt], XmNsliderSize, &size); cnt++;
    XtGetValues(tw->h_scrollb, args, cnt);

    tw->cur_col = (int)((float) cb->value / (float)(max - size) * 
						(float)(tw->cols - npl));
    
    if (tw->cur_col > tw->cols - npl)
      tw->cur_col = tw->cols - npl;
    break;
  }

  if (oldmargin != tw->cur_col) 			/* nothing changed */
    treeHmove(tw);
}

/* treeHmove: ***************************************************************
 * move the button plane so that the current row appears at the top 
 ****************************************************************************/

static void treeHmove(tw)
     treeWindow *tw;
{
  XtMoveWidget(tw->plane, 
	       -tw->cur_col * (DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP),
	       -tw->cur_row * (DIRBUTTONHEIGHT(tw->font) + DIRBUTTONVGAP));
 
  XDrawSegments(XtDisplay(Head->root), XtWindow(tw->plane), Head->fm_gc, 
				tw->connections, tw->conCnt);
}

/* ARGSUSED */
void SelectTreeWnd(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String params;
     int *num_params;
{
  Arg args[1];
  int cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &curTreeWnd); cnt++;
  XtGetValues(w, args, cnt);
}

/* nodesPerRow: *************************************************************
 * calculate the number of node-lines that fit in the tree window
 ****************************************************************************/

nodesPerRow(tw)
     treeWindow *tw;
{
  unsigned int w, h, bw, d;
  int x, y;
  Window rw;

  XGetGeometry(XtDisplay(Head->root), XtWindow(tw->work), &rw, &x, &y, &w, &h, 
	       		&bw, &d);

  return(ROUNDOFF(h, (DIRBUTTONHEIGHT(tw->font) + DIRBUTTONVGAP)));
}

/* nodesPerLine: *************************************************************
 * calculate the number of node-columns that fit in the tree window 
 ****************************************************************************/
 
nodesPerLine(tw)
     treeWindow *tw;
{
  unsigned int w, h, bw, d;
  int x, y;
  Window rw;

  XGetGeometry(XtDisplay(Head->root), XtWindow(tw->work), &rw, &x, &y, &w, &h, 
	       		&bw, &d);

  return(ROUNDOFF(w, (DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP)));
}

/* confTreeWindow: ***********************************************************
 * callback function that is called when the viewport of the tree window has 
 * been configured. If the size has been changed this function must set new
 * scrolbar values if any to reflect the actual situation.
 ****************************************************************************/
/*ARGSUSED*/
void confTreeWindow(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String params;
     int *num_params;
{
  treeWindow *tw;
  Arg args[1];
  int cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &tw); cnt++;
  XtGetValues(w, args, cnt);

  if (! tw )
    return;
  if (tw->v_scrollb)
    setTree_vScrollbar(tw, event->xconfigure.height);
  if (tw->h_scrollb)
    setTree_hScrollbar(tw, event->xconfigure.width);
}

/* setTree_vScrollbar: *******************************************************
 * calculates new value and size for the vertical scrollbar 
 ****************************************************************************/

static void setTree_vScrollbar(tw, height)
     treeWindow *tw;
     int height;
{
  int size, inc, pageinc, cnt, npr, value;
  Arg args[6];

  npr = ROUNDOFF(height, (DIRBUTTONHEIGHT(tw->font) + DIRBUTTONVGAP));
  if (npr >= tw->rows) {
    size = height;
    value = 0;
    inc = 1;
    pageinc = 1;
    tw->flags &= ~VSCROLL;
    if (tw->cur_row != 0) {
      tw->cur_row = 0;
      treeVmove(tw);
    }
  } else {
    if (tw->cur_row + npr > tw->rows) {		/* adjust rows */
      tw->cur_row = tw->rows - npr;
      treeVmove(tw);
    }
    tw->flags |= VSCROLL;
    size = (int) ((float) npr / (float) tw->rows * (float) height);
    inc = (int)((float) (height - size) / (float)(tw->rows - npr));
    pageinc = 
      (int) ((float)(height - size) / (float) MIN(4, tw->rows - npr));
    value = tw->cur_row * inc;
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNsliderSize, size); cnt++;
  XtSetArg(args[cnt], XmNmaximum, height); cnt++;
  XtSetArg(args[cnt], XmNincrement, inc); cnt++;
  XtSetArg(args[cnt], XmNpageIncrement, pageinc); cnt++;
  XtSetArg(args[cnt], XmNvalue, value); cnt++;
  XtSetValues(tw->v_scrollb, args, cnt);
}

/* setTree_hScrollbar: *******************************************************
 * calculates new value and size for the horizontal scrollbar 
 ****************************************************************************/

static void setTree_hScrollbar(tw, width)
     treeWindow *tw;
     int width;
{
  int size, inc, pageinc, cnt, npl, value;
  Arg args[6];

  npl = ROUNDOFF(width, (DIRBUTTONWIDTH(tw->font) + DIRBUTTONHGAP));

  if (npl >= tw->cols) {
    size = width;
    value = 0;
    inc = 1;
    pageinc = 1;
    tw->flags &= ~HSCROLL;
    if (tw->cur_col != 0) {
	tw->cur_col = 0;
	treeHmove(tw);
      }
  } else {
    if (tw->cur_col + npl > tw->cols) {		/* adjust cols */
      tw->cur_col = tw->cols - npl;
      treeHmove(tw);
    }
    tw->flags |= HSCROLL;
    size = (int) ((float) npl / (float) tw->cols * (float) width);
    inc = (int) ((float)(width - size) / (float)(tw->cols - npl));
    pageinc = 
      (int) ((float) (width - size) / (float) MIN(4, tw->cols - npl));
    value = tw->cur_col * inc;
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNsliderSize, size); cnt++;
  XtSetArg(args[cnt], XmNmaximum, width); cnt++;
  XtSetArg(args[cnt], XmNincrement, inc); cnt++;
  XtSetArg(args[cnt], XmNpageIncrement, pageinc); cnt++;
  XtSetArg(args[cnt], XmNvalue, value); cnt++;
  XtSetValues(tw->h_scrollb, args, cnt);
}
