/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"

#include <sys/dir.h>
#include <strings.h>

static char SccsID[] = "@(#)directory.c	1.3\t1/27/93";            /* sccs keyword */

static char msg[200];

dirWindow *allocateDirWindow();
dirItem *allocateDirItem();
dirWindow *lookupWindowByName();
dirWindow *lookifDisplayed();

/* listDirectory: ************************************************************
 * display the selected directories in their own directory windows
 ****************************************************************************/

void listDirectory(wnd, what)
     dirWindow *wnd;
     int what;
{
  register dirItem *iptr;
  static char path[PATHLEN],upath[PATHLEN];
  dirItem *list;
  int cnt;
  dirWindow *dirWnd = (dirWindow *) NULL;
  register char *cptr, *nptr;
  char buf[PATHLEN];

  switch (what) {
  case DIR_FATHER:			/* goto father directory */
    if (strcmp(wnd->u_path,"/") == 0) {
      exAlert (Head->root, 100, 100, 
	       "Cannot go to father directory: alread at root");
      break;
    }
    if ((cptr = rindex(wnd->u_path, '/')) == (char *) NULL) {
      sprintf(msg, "Cannot find father directory: %s", wnd->path);
      exAlert(Head->root, 100, 100, msg);
      break;
    }

    strcpy(upath, wnd->u_path);		/* strip off last component */
    nptr = rindex(upath, '/');
    if (nptr == upath) 
      *++nptr = EOS;
    else
      *nptr++ = EOS;

    if (lookifDisplayed(upath))		/* terminate, if already displayed */
      break;

    strcpy(path, wnd->path);
    cptr = rindex(path, '/');
    if (cptr == path) 
      *++cptr = EOS;
    else
      *cptr = EOS;

    if ((cnt = readDirectory(upath, &list)) == -1)
      break;
    
    dirWnd = allocateDirWindow(path, upath, list, cnt);
    sortItemList(dirWnd);
    markVisibleItems(dirWnd, wnd->filePattern, False);
    displayDirWindow(dirWnd);

    dirWnd->next = (dirWindow *) NULL;
    Head->last->next = dirWnd;
    dirWnd->prev = Head->last;
    Head->last = dirWnd;
    
    /* mark the displayed item in the newly displayed dir-window */
    for (iptr = dirWnd->itemList; iptr; iptr = iptr->next) {
      SET_ITEM_STATE(dirWnd, iptr);
      if (strcmp(iptr->D_name, nptr) == EQUAL) {
	iptr->flags |= ITEM_DISPLAYED;
	fmdisplayItem(iptr->dirWnd, iptr, True);
	break;
      }
    }

    checkParentWindow(upath, True);	/* mark father if displayed */
    updateTree(upath, True);		/* hilite dir. in displayed trees */
    break;
  case DIR_ANY:				/* let user input directory name */
    nptr = exGetFilename(Head->root, -1, -1, 
			 "Enter directory: ", wnd->u_path, 50);
    if (! nptr)
      break;

    if (*nptr == '\n' || *nptr == '\r' || *nptr == EOS)
      break;

    cleanPath(path, nptr, 50);
    free(nptr);
    if (path[strlen(path) - 1] == '/')
      path[strlen(path) - 1] = EOS;

    if (lookifDisplayed(path))
      break;

    if ((cnt = readDirectory(path, &list)) == -1)
      break;
    
    dirWnd = allocateDirWindow(path, path, list, cnt);
    sortItemList(dirWnd);
    markVisibleItems(dirWnd, wnd->filePattern, False);
    displayDirWindow(dirWnd);

    dirWnd->next = (dirWindow *) NULL;
    Head->last->next = dirWnd;
    dirWnd->prev = Head->last;
    Head->last = dirWnd;

    checkParentWindow(path, True); 	/* mark father if displayed */
    updateTree(path, True);		/* hilite dir. in displayed trees */
    
    break; 
  case DIR_SELECTED:		/* list the current selected directory */
    for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
      SET_ITEM_STATE(wnd, iptr);
      if (iptr->flags & ITEM_SELECTED && iptr->flags & ITEM_VISIBLE) {
	if (iptr->flags & ITEM_DISPLAYED)   	/* item already displayed */
	  continue;

	strcpy(buf, wnd->u_path);
	strcat(buf,"/");
	strcat(buf, iptr->D_name);
	cleanPath(upath, buf, PATHLEN);

	if (upath[strlen(upath) - 1] == '/')
	  upath[strlen(upath) - 1] = EOS;

	strcpy(buf, wnd->path);
	strcat(buf, "/");
	strcat(buf, ITEM_NAME(iptr));
	cleanPath(path, buf, PATHLEN);

	if (path[strlen(path) - 1] == '/')
	  path[strlen(path) - 1] = EOS;

	if ((cnt = readDirectory(upath, &list)) == -1)
	  continue;
	
	dirWnd = allocateDirWindow(path, upath, list, cnt);
	sortItemList(dirWnd);
	markVisibleItems(dirWnd, wnd->filePattern, False);
	displayDirWindow(dirWnd);
	
	dirWnd->next = (dirWindow *) NULL;
	Head->last->next = dirWnd;
	dirWnd->prev = Head->last;
	Head->last = dirWnd;
	
	updateTree(upath, True);		/* hilite dir. in trees */
	
	iptr->flags &= ~ITEM_SELECTED;          /* deselect item */
	iptr->flags |= ITEM_DISPLAYED;	       	/* mark item displayed */
	SET_MENU_STATE(wnd, iptr);		/* mark menu item unselected */
	fmdisplayItem(iptr->dirWnd, iptr, True);/* display item */
      }
    }
  }
}

/* editDirectory: ************************************************************
 * call the interactive attribute editor for each selected file 
 ****************************************************************************/

/*ARGSUSED*/
void editDirectory(wnd)
     dirWindow *wnd;
{
  register dirItem *iptr;
  char **file_list;
  int i;

  for (i = 0, iptr = wnd->itemList; iptr; iptr = iptr->next) {
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & ITEM_SELECTED) {
      if (i % 10 == 0) 
	if (i == 0)
	  file_list = (char **) calloc(10, sizeof(char *));
	else
	  file_list = (char **) realloc(file_list, (i + 10) * sizeof(char *));

      file_list[i++] = strCopy(iptr->D_name);
    }  
  }
  openAed(Head->aed_form, wnd->u_path, file_list, i);
  aedInfo(Head->aed_form, 0);
}

/* updateDirWindow: ********************************************************** 
 * the contents of a directory has been changed after the corresponding 
 * directory-window has been constructed. In this case the directory is 
 * re-read and the item list updated. If a whole directory is deleted, that 
 * is displayed, the corresponding directory-window is deleted.
 *****************************************************************************/

updateDirWindow(w)
     dirWindow *w;
{
  int cc, cnt;
  dirItem *newlist, *oldlist;
  register dirItem *iptr, *optr;
  dirWindow *dirw;
  Boolean rflag = False;			/* if true, remake buttons */

  if ((cnt = readDirectory(w->u_path, &newlist)) <= 0) {
    exAlert(Head->root, 100, 100, "No matching files detected !");
    return;
  }
  
  w->itemCnt = cnt; 
  setReadTime(w);
  oldlist = w->itemList;
  w->itemList = newlist;
  sortItemList(w);
  
  /* assign the newly read items the same flags; if an item has been deleted,
   * look if it is a directory that is already displayed; if so, delete the 
   * corresponding directory window 
   */

  optr = oldlist;
  iptr = w->itemList;
  
  while(1) {					/* loop through items */
    cc = strcmp(iptr->D_name, optr->D_name);
    if (cc == EQUAL) { 				/* same item: supply values */
      iptr->a_name = strCopy(optr->a_name);
      iptr->flags  = optr->flags;
      iptr->slot = optr->slot;
      iptr = iptr->next;
      optr = optr->next;
    } else if (cc > 0) {  			/* item in old-list deleted */
      rflag = True;
      if (optr->flags & ITEM_DISPLAYED) { 	/*  dirwindow displayed */
	if ((w = lookupWindowByName(ITEM_NAME(optr))) != (dirWindow *) NULL)
	  removeDirWindow(w);
	else {
	  sprintf(msg, "Internal error: cannot locate window -> %s", 
		  					ITEM_NAME(optr));
	  exError("fm", msg, NULL);
	  return;
	}
      }
      optr = optr->next;
    } else {					/* new item */
      iptr = iptr->next;
      rflag = True;
    }
   
    if (! optr) {				/* orig. list exhausted */
      rflag = True;			      	/* generate new buttons */
      break;
    }
    
    if (! iptr) {				/* rem. items are deleted */
      rflag = True;
      for (; optr; optr = optr->next) {
	if (optr->flags & ITEM_DISPLAYED) { 	/*  window displayed */
	  if (dirw = lookupWindowByName(ITEM_NAME(optr)))
	    removeDirWindow(dirw);
	  else {
	    sprintf(msg, "Internal error: cannot locate window -> %s ",
		    					ITEM_NAME(optr));
	    exAlert(Head->root, 100, 100, msg);
	    return;
	  }
	}
      }
      break;					/* done with both lists */
    }
  }

  freeItems(oldlist);

  for (iptr = w->itemList; iptr; iptr = iptr->next)
    iptr->dirWnd = w;

  if (rflag) 
    markVisibleItems(w, w->filePattern, True);

  setDirBoxSizeHints(w);			/* adapt window size */
}

/* listDir: *****************************************************************
 * display the directory given by path in a directory window 
 ****************************************************************************/

listDir(wnd, path)
     dirWindow *wnd;
     char *path;
{
  dirItem *list;
  int cnt;
  dirWindow *dirWnd;

  if ((cnt = readDirectory(path, &list)) == -1)
    return;
  dirWnd = allocateDirWindow(path, path, list, cnt);
  sortItemList(dirWnd);
  markVisibleItems(dirWnd, wnd->filePattern, False);
  displayDirWindow(dirWnd);

  /* chain window */
  dirWnd->next = (dirWindow *) NULL;
  Head->last->next = dirWnd;
  dirWnd->prev = Head->last;
  Head->last = dirWnd;

  checkParentWindow(path, True);	/* mark dir. in displayed windows */
  updateTree(path, True);		/* hilite dir. in displayed trees */
}
/* readDirectory: ************************************************************
 * reads a diretory as specified by the path-parameter; it builds a list 
 * of item-entries, one for each file in the specified directory. The list is 
 * returned in the 'list'-parameter. The function returns the number of 
 * items.
 *****************************************************************************/

int readDirectory(path, list)
     char *path;
     dirItem **list;		/* head of the item list that is returned */
{
  struct direct *d_entry;
  DIR *dirp;
  register dirItem *ptr;
  dirItem *new_dirItem();
  dirItem *itemHead = NULL;
  int icnt;
  
  if (! isDirFile(path)) {
    sprintf(msg, "File is not a directory: %s",path);
    exAlert(Head->root, 100, 100, msg);
    return(-1);
  }
  
  if ((dirp = opendir(path)) == NULL) {
    sprintf(msg, "Cannot open directory: %s",path);
    exAlert(Head->root, 100, 100, msg);
    return(-1);
  }
  
  for (d_entry = readdir(dirp), icnt = 0; d_entry != NULL; 
                                        d_entry = readdir(dirp), icnt++) {
    if ((ptr = allocateDirItem()) == NULL) 
      fatal("Cannot allocate items","readDirectory");
    
    ptr->flags = 0;

    ptr->next = itemHead;
    itemHead = ptr;
    ptr->D_name = strCopy(d_entry->d_name);
    getStat(ptr, path, ptr->D_name);
    
    if (isdirItem(ptr))
      ptr->flags |= ITEM_ISDIR;
  }
  closedir(dirp);
  *list = itemHead;
  return(icnt); 
}

/* needUpdate: ***************************************************************
 * compares the modification time of the associated directory that has been 
 * stored when the directory-window has been last updated (first when it has 
 * been created) with the time when the directory-window has been last updated.
 * If the directory has been modified after the dirrectory window has been 
 * created (the directories status-change time is greater then the directory-
 * windows readTime, the  directory-window must be updated.
 ****************************************************************************/

needUpdate(w)
     dirWindow *w;
{
  return(getStatusChangeTime(w->u_path) > w->readTime);
}




