/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <strings.h>
#include <string.h>
#include <exBasics.h>
#include <exEdit.h>
#include <exMenu.h>

static char BUFFER0[EX_MAXSTR], BUFFER1[EX_MAXSTR];
static char CWD[512];

/* expandFileName: ***********************************************************
 * This function provides the means to request the expansion of a filename.
 *****************************************************************************/

Boolean expandFileName(info, multiple, show_menu)
     exExpandInfo *info; 
     Boolean multiple;
     Boolean show_menu;
{
  int rc, i;
  char *oldbuf, *pattern, *str, *ptr, **matchList = NULL, *prefix = NULL;
  Boolean copyPath = False;
  MHmenu *mptr;
  exMenu *menu;

  info->retBuf = NULL;
  info->matchCnt = 0;

  strcpy(BUFFER1, info->word);
  oldbuf = strCopy(BUFFER1);

  if (BUFFER1[0] != '/' && BUFFER1[0] != '~') {	/* relative path */
    if (! (* info->getcwd)(CWD, sizeof(CWD))) {
      return(False);
    }
    strcpy(BUFFER0, CWD);
    strcat(BUFFER0, "/");
    strcat(BUFFER0, BUFFER1);
  } else 
    strcpy(BUFFER0, BUFFER1);

  if (strchr(BUFFER0, '.')) {	
    cleanPath(BUFFER1, BUFFER0, EX_MAXSTR);	/* rem interm '.' and '..' */
    if (oldbuf[strlen(oldbuf) - 1] != '/')
      BUFFER1[strlen(BUFFER1) - 1] = EOS;    	/* rem trailing '/' */
    if (strcmp(oldbuf, BUFFER1))
      copyPath = True;				/* replace entire path */
  } else
    strcpy(BUFFER1, BUFFER0);

  /* generate directory and pattern */

  if (strpbrk(BUFFER1, "~$"))			/* to be expanded */
    copyPath = True;				/* replace entire path */
  pattern =  (char *) malloc(strlen(BUFFER1) + 2);

  if (BUFFER1[0] == '~') {			/* ~ expansion */
    if (! strchr(BUFFER1, '/'))	{		/* expand ~<user> */ 
      strcpy(pattern, BUFFER1);
      strcat(pattern, "/");
    } else {					/* expand ~/.. */
      strcpy(pattern, BUFFER1);
      strcat(pattern, "*");
    }
  } else if (BUFFER1[0] == '/') {		/* expand abs. path */
    strcpy(pattern, BUFFER1);
    strcat(pattern, "*");
  } else {					/* expand rel. path */
    strcpy(pattern, BUFFER1);
    strcat(pattern, "*");
  }

  BUFFER1[0] = EOS;
  BUFFER0[0] = EOS;

  rc = exParsePath(pattern, &prefix, BUFFER1, sizeof(BUFFER1));
  free(pattern);
  
  if (rc < 1) {
    free(oldbuf);
    if (prefix)
      free(prefix);
    return(False);
  }

  if (prefix) {					/* new prefix */
    strcpy(BUFFER0, prefix);
    strcat(BUFFER0, "/");
  }
  
  info->matchCnt = rc;

  if (rc == 1) {				/* one expansion */
    for (ptr = &BUFFER1[strlen(BUFFER1) - 1]; 
	 *ptr == ' ' && ptr > BUFFER1; ptr--)
      *ptr = EOS;
    
    if (copyPath) {				/* fill in entire path */
      if (BUFFER0[strlen(BUFFER0) - 1] != '/')
	strcat(BUFFER0, "/");

      if (BUFFER1[0] == '/')
	strcat(BUFFER0, &BUFFER1[1]);
      else 
	strcat(BUFFER0, BUFFER1);

      if (isDirFile(BUFFER0))
	strcat(BUFFER0, "/");
      info->retBuf = strCopy(BUFFER0);
    } else {					/* leave old path */
      if (ptr = strrchr(oldbuf, '/')) {		/* a path has been expanded */	
	*++ptr = EOS;				/* strip off last component */
	strcpy(BUFFER0, oldbuf);
	strcat(BUFFER0, BUFFER1);

	if (BUFFER0[0] != '/')
	  sprintf(BUFFER1,"%s/%s",CWD,BUFFER0);	/* generate abs. path */
	else
	  strcpy(BUFFER1, BUFFER0);
	if (isDirFile(BUFFER1))			/* if directory ... */
	  strcat(BUFFER0, "/");			/* ... append '/' */
	info->retBuf = strCopy(BUFFER0);
	return(True);
      } else {					/* a file has been expanded */
	strcat(BUFFER0, BUFFER1);
	if (isDirFile(BUFFER0))
	  strcat(BUFFER1, "/");
	info->retBuf = strCopy(BUFFER1);	/* no path, fill in name */
      }
    }
  } else {					/* multiple expansions */
    if (multiple) {				/* multiple insert allowed */
      matchList = Calloc(rc, sizeof(char *));
      (void) makeStringList(BUFFER1, matchList, rc, ' ');
      sortStringList(matchList, rc);
      
      if (! copyPath) {				/* use path provided by esh */
	if (ptr = strrchr(info->word, '/')) {	/* strip of last component */
	  *++ptr = EOS;
	  strcpy(BUFFER0, info->word);
	} else  	
	  BUFFER0[0] = EOS;
      } else { 					/* use generated path */
	if (BUFFER0[strlen(BUFFER0) - 1] != '/')
	  strcat(BUFFER0, "/");
      }
      
      ptr = &BUFFER0[strlen(BUFFER0)];
      BUFFER1[0] = EOS;
      for (i = 0; i < rc; i++) {
	strcpy(ptr, matchList[i]);
	if (isDirFile(BUFFER0))
	  strcat(ptr, "/");
	
	strcat(BUFFER1, BUFFER0);
	strcat(BUFFER1, " ");
	free(matchList[i]);
      }
      info->retBuf = strCopy(BUFFER1);
    } else {					/* multi-insert not allowed */
      matchList = Calloc(rc, sizeof(char *));
      (void) makeStringList(BUFFER1, matchList, rc, ' ');
      sortStringList(matchList, rc);
      if (show_menu) {				/* let user select by menu */
	setWorkApplication();

	if (! (mptr = buildMenuFromStrings(prefix ? prefix : "Completions", 
				   MDS_L_POPUP, matchList, matchList, rc))) {
	  if (prefix)
	    free(prefix);
	  return(False);
	}

	for (i = 0; i < rc; i++)
	  free(matchList[i]);

	setSelCnt(mptr, 0);
	setSelType(mptr, MDS_CNT_GT);
	setMaxItem(mptr, 10);
	setMenuPos(mptr, 5, 5, MDS_POS_CREL);

	if (! (menu = exMakeMenu(NULL, NULL, info->parent, mptr)))
	  return(False);

	exCallMenu(menu);
	str = strCopy(exGetMenuReturnValue(menu));
	freeMenu(menu);
	deleteWorkMenu(mptr);

	if (str) {				/* items selected in menu */
	  rc = makeStringList(str, matchList, rc, ' ');
	  if (! copyPath) {			/* use path provided by esh */
	    if (ptr = strrchr(info->word, '/')) {/* strip off last component */
	      *++ptr = EOS;
	      strcpy(BUFFER0, info->word);
	    } else  	
	      BUFFER0[0] = EOS;
	  } else { 				/* use generated path */
	    if (BUFFER0[strlen(BUFFER0) - 1] != '/')
	      strcat(BUFFER0, "/");
	  }
	  
	  ptr = &BUFFER0[strlen(BUFFER0)];
	  for (i = 0; i < rc; i++) {
	    strcpy(ptr, matchList[i]);
	    if (isDirFile(BUFFER0))
	      strcat(ptr, "/");
	    
	    if (i == 0) 
	      strcpy(BUFFER1, BUFFER0);
	    else {
	      strcat(BUFFER1, " ");
	      strcat(BUFFER1, BUFFER0);
	    }
	    free(matchList[i]);
	  }
	  free(matchList);
	  free(str);
	  info->retBuf = strCopy(BUFFER1);
	  info->matchCnt = rc;
	} 
      } else {					/* provide common prefix */
	register j, k;
	Boolean goon = True;
	for (k = 0; goon; k++) {
	  for (j = 1; j < rc; j++) {
	    if (matchList[0][k] != matchList[j][k]) {
	      goon = False;
	      if (! matchList[0][k]) 
		strcpy(BUFFER1, matchList[0]);
	      else {
		matchList[0][k] = EOS;
		strcpy(BUFFER1, matchList[0]);
	      }		

	      for (k = 0; k < rc; k++)
		free(matchList[k]);
	      
	      if (copyPath) {			/* fill in entire path */
		if (BUFFER0[strlen(BUFFER0) - 1] != '/')
		  strcat(BUFFER0, "/");
		
		if (BUFFER1[0] == '/')
		  strcat(BUFFER0, &BUFFER1[1]);
		else 
		  strcat(BUFFER0, BUFFER1);
		info->retBuf = strCopy(BUFFER0);
	      } else {				/* leave old path */
		if (ptr = strrchr(oldbuf,'/')) {/* a path has been expanded */	
		  *++ptr = EOS;			/* strip off last component */
		  strcpy(BUFFER0, oldbuf);
		  strcat(BUFFER0, BUFFER1);
		  
		  if (BUFFER0[0] != '/')	/* generate abs. path */
		    sprintf(BUFFER1,"%s/%s",CWD,BUFFER0);
		  else
		    strcpy(BUFFER1, BUFFER0);
		  info->retBuf = strCopy(BUFFER0);
		  return(True);
		} else {			/* a file has been expanded */
		  strcat(BUFFER0, BUFFER1);
		  info->retBuf = strCopy(BUFFER1);/* no path, fill in name */
		}
	      }
	    }
	    break;
	  }
	}
      }
    }
  }
  
  if (prefix)
    free(prefix);

  if (matchList)
    free(matchList);

  free(oldbuf);

  return(True);
}

/* setExpandInfo: ************************************************************
 * auxiliary function to set the given components in the info structure needed
 * by the function for file name expansion
 ****************************************************************************/

exExpandInfo *setExpandInfo(parent, word, getcwd)
     Widget parent;
     char *word;
     int (* getcwd) ();
{
  register exExpandInfo *info;

  if (! (info = (exExpandInfo *) Malloc(sizeof(exExpandInfo))))
    return(NULL);

  info->parent = parent;
  info->getcwd = getcwd;
  info->word = word;

  return(info);
}
